#ifndef _ASM_PPC_BITOPS_H_
#define _ASM_PPC_BITOPS_H_

/*
 * For the benefit of those who are trying to port Linux to another
 * architecture, here are some C-language equivalents.  You should
 * recode these in the native assembly language, if at all possible.
 * To guarantee atomicity, these routines call cli() and sti() to
 * disable interrupts while they operate.  (You have to provide inline
 * routines to cli() and sti().)
 *
 * Also note, these routines assume that you have 32 bit integers.
 * You will have to change this if you are trying to port Linux to the
 * Alpha architecture or to a Cray.  :-)
 * 
 * C language equivalents written by Theodore Ts'o, 9/26/92
 */

#include "asm/system.h"  /* For cli/sti declaration */

extern unsigned long ffz(unsigned long);

#define BIT(n) 1<<(n&0x1F)
typedef unsigned long BITFIELD;

extern __inline__ int set_bit(int nr, void * add)
{
       int	mask, oldbit;
  BITFIELD *addr = add;
       
	int s = _disable_interrupts();
	addr += nr >> 5;
	mask = BIT(nr);
	oldbit = (mask & *addr) != 0;
	*addr |= mask;
	_enable_interrupts(s);

	
	return oldbit;
}

extern __inline__ int change_bit(int nr, void *add)
{
  	BITFIELD *addr = add;
	int	mask, retval;
	int s = _disable_interrupts();
	addr += nr >> 5;
	mask = BIT(nr);
	retval = (mask & *addr) != 0;
	*addr ^= mask;
	_enable_interrupts(s);
	return retval;
}

extern __inline__ int clear_bit(int nr, void *add)
{
        BITFIELD *addr = add;
	int	mask, retval;
	int s = _disable_interrupts();
	addr += nr >> 5;
	mask = BIT(nr);
	retval = (mask & *addr) != 0;
	*addr &= ~mask;
	_enable_interrupts(s);
	return retval;
}

extern __inline__ int test_bit(int nr, const void *add)
{
	int	mask, retval;
	const BITFIELD *addr = (const BITFIELD *)add;
	int s = _disable_interrupts();

	addr += nr >> 5;
	mask = BIT(nr);
	retval = ((mask & *addr) != 0);
	_enable_interrupts(s);
	return retval;
}

#if 0
extern __inline__ int find_first_zero_bit(void *add, int len)
{
	int	mask, nr, i;
	BITFIELD *addr = add;
	nr = 0;
	while (len)
	{
		if (~*addr != 0)
		{ /* Contains at least one zero */
			for (i = 0;  i < 32;  i++, nr++)
			{
				mask = BIT(nr);
				if ((mask & *addr) == 0)
				{
					return (nr);
				}
			}
		}
		len -= 32;
		addr++;
		nr += 32;
	}
	return (0);  /* Shouldn't happen */
}

extern __inline__ int find_next_zero_bit(void *add, int len, int nr)
{
	int	mask, i;
	BITFIELD *addr = add;
	addr += nr >> 5;
	len -= nr;
	while (len)
	{
		if (*addr != 0xFFFFFFFF)
		{ /* Contains at least one zero */
			for (i = 0;  i < 32;  i++, nr++)
			{
				mask = BIT(nr);
				if ((mask & *addr) == 0)
				{
printk("Bit: %d(%d), Pat: %x\n", nr, nr&0x1F, *addr);					
					return (nr);
				}
			}
		}
		len -= 32;
		addr++;
		nr += 32;
	}
	return (0);  /* Shouldn't happen */
}
#endif

extern __inline__ int __ext2_set_bit(int nr, void *addr)
{
	int		mask, old_value;
	unsigned char	*ADDR = (unsigned char *) addr;

	ADDR += nr >> 3;
	mask = 1 << (nr & 0x07);
	old_value = *ADDR & mask;
	*ADDR |= mask;
	return (old_value);
}

extern __inline__ int __ext2_clear_bit(int nr, void *addr)
{
	int		mask, old_value;
	unsigned char	*ADDR = (unsigned char *) addr;

	ADDR += nr >> 3;
	mask = 1 << (nr & 0x07);
	old_value = *ADDR & mask;
	*ADDR &= ~mask;
	return (old_value);
}

extern __inline__ int __ext2_test_bit(int nr, __const__ void * addr)
{
	int			mask;
	__const__ unsigned char	*ADDR = (__const__ unsigned char *) addr;

	ADDR += nr >> 3;
	mask = 1 << (nr & 0x07);
	return ((mask & *ADDR) != 0);
}

extern __inline__ unsigned short __swab16(unsigned short value)
{
        return ((value >> 8) | (value << 8));
}     

extern __inline__ unsigned long __swab32(unsigned long value)
{
        return ((value >> 24) | ((value >> 8) & 0xff00) |
               ((value << 8) & 0xff0000) | (value << 24));
}     

#define __ext2_find_first_zero_bit(addr, size) \
        __ext2_find_next_zero_bit((addr), (size), 0)

extern __inline__ unsigned long __ext2_find_next_zero_bit(void *addr, unsigned long size, unsigned long offset)
{
	unsigned long *p = ((unsigned long *) addr) + (offset >> 5);
	unsigned long result = offset & ~31UL;
	unsigned long tmp;

	if (offset >= size)
		return size;
	size -= result;
	offset &= 31UL;
	if(offset) {
		tmp = *(p++);
		tmp |= __swab32(~0UL >> (32-offset));
		if(size < 32)
			goto found_first;
		if(~tmp)
			goto found_middle;
		size -= 32;
		result += 32;
	}
	while(size & ~31UL) {
		if(~(tmp = *(p++)))
			goto found_middle;
		result += 32;
		size -= 32;
	}
	if(!size)
		return result;
	tmp = *p;

found_first:
	return result + ffz(__swab32(tmp) | (~0UL << size));
found_middle:
	return result + ffz(__swab32(tmp));
}

#endif /* _ASM_PPC_BITOPS_H */


